(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[4648],{12037:function(i,e,t){(window.__NEXT_P=window.__NEXT_P||[]).push(["/consent/privacy-notices/[id]",function(){return t(14755)}])},14755:function(i,e,t){"use strict";t.r(e);var n=t(24246),c=t(96013),s=t(86677),r=t(43124),a=t(87162),o=t(37541),d=t(40793),u=t(26796);e.default=()=>{let i=(0,s.useRouter)(),e="";i.query.id&&(e=Array.isArray(i.query.id)?i.query.id[0]:i.query.id);let{data:t,isLoading:l}=(0,d._g)(e),{data:h}=(0,d.Wq)(e);return l?(0,n.jsx)(r.Z,{title:"Privacy notice",children:(0,n.jsx)(c.M5Y,{children:(0,n.jsx)(c.$jN,{})})}):t?(0,n.jsxs)(r.Z,{title:"Privacy notice ".concat(t.name),children:[(0,n.jsx)(o.Z,{heading:"Privacy Notices",breadcrumbItems:[{title:"All privacy Notices",href:a.IT},{title:t.name}]}),(0,n.jsxs)(c.xuv,{width:{base:"100%",lg:"70%"},"data-testid":"privacy-notice-detail-page",children:[(0,n.jsx)(c.xvT,{fontSize:"sm",mb:8,children:"Configure your privacy notice including consent mechanism, associated data uses and the locations in which this should be displayed to users."}),(0,n.jsx)(u.Z,{privacyNotice:t,availableTranslations:h})]})]}):(0,n.jsx)(r.Z,{title:"Privacy notice",children:(0,n.jsxs)(c.xvT,{children:["No privacy notice with id ",e," found."]})})}}},function(i){i.O(0,[4554,2866,9111,8837,3528,1241,2163,2888,9774,179],function(){return i(i.s=12037)}),_N_E=i.O()}]);