"""adds additional config vals on experience config

Revision ID: fc82ab64bd5e
Revises: a6d9cdfcc7dc
Create Date: 2024-07-29 14:55:01.609732

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "fc82ab64bd5e"
down_revision = "a6d9cdfcc7dc"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "experienceconfigtemplate",
        sa.Column("show_layer1_notices", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "experienceconfigtemplate",
        sa.Column(
            "layer1_button_options",
            sa.String(),
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("show_layer1_notices", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfig",
        sa.Column(
            "layer1_button_options",
            sa.String(),
            nullable=True,
        ),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("show_layer1_notices", sa.Boolean(), nullable=True),
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column(
            "layer1_button_options",
            sa.String(),
            nullable=True,
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("privacyexperienceconfighistory", "layer1_button_options")
    op.drop_column("privacyexperienceconfighistory", "show_layer1_notices")
    op.drop_column("privacyexperienceconfig", "layer1_button_options")
    op.drop_column("privacyexperienceconfig", "show_layer1_notices")
    op.drop_column("experienceconfigtemplate", "layer1_button_options")
    op.drop_column("experienceconfigtemplate", "show_layer1_notices")
    # ### end Alembic commands ###
