(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[4745],{37441:function(e,t,n){(window.__NEXT_P=window.__NEXT_P||[]).push(["/data-catalog/[systemId]/resources/[resourceUrn]",function(){return n(4635)}])},4635:function(e,t,n){"use strict";n.r(t);var r=n(24246),s=n(86677),c=n(58665),u=n(43124),o=n(87162),a=n(37541),l=n(67219),i=n(71819),d=n(22153);t.default=()=>{var e;let{query:t}=(0,s.useRouter)(),n=t.systemId,_=t.resourceUrn,{data:f,isLoading:m}=(0,d.rn)(n),h=(0,s.useRouter)(),k=(0,i.pN)(_,"".concat(o.mX,"/").concat(n,"/resources"));return m?(0,r.jsx)(c.Z,{}):(0,r.jsxs)(u.Z,{title:"Data catalog",children:[(0,r.jsx)(a.Z,{heading:"Data catalog",breadcrumbItems:[{title:"All systems",href:o.mX},{title:null!==(e=null==f?void 0:f.name)&&void 0!==e?e:null==f?void 0:f.fides_key,href:"".concat(o.mX)},...k]}),(0,r.jsx)(l.Z,{resourceUrn:_,onRowClick:e=>h.push("".concat(o.mX,"/").concat(f.fides_key,"/resources/").concat(e.urn))})]})}}},function(e){e.O(0,[8033,5559,4554,2866,9278,8540,5632,7548,1827,2888,9774,179],function(){return e(e.s=37441)}),_N_E=e.O()}]);