.CalloutNavCard_container__DXPJb{border-radius:6px;display:flex;flex-grow:1;transition:box-shadow .2s ease-in-out}.CalloutNavCard_container__DXPJb:hover{box-shadow:0 1px 2px 0 rgba(0,0,0,.06),0 1px 3px 0 rgba(0,0,0,.1)}.CalloutNavCard_card__q_XX6{flex-grow:1;border-radius:0 6px 6px 0;border-left:none;background-color:var(--fidesui-corinth)}