(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[4727],{30410:function(n,_,u){(window.__NEXT_P=window.__NEXT_P||[]).push(["/login/[provider]",function(){return u(44217)}])}},function(n){n.O(0,[2888,9774,179],function(){return n(n.s=30410)}),_N_E=n.O()}]);