(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[3691],{8828:function(e,t,i){(window.__NEXT_P=window.__NEXT_P||[]).push(["/systems/configure/[id]",function(){return i(13216)}])},11574:function(e,t,i){"use strict";var n=i(24246),s=i(96013),r=i(36848);t.Z=e=>{let{data:t,border:i="partial",borderWidth:a=2,...l}=e;return(0,n.jsx)(s.mQc,{colorScheme:"complimentary",...l,children:(0,n.jsx)(s.tdY,{width:"partial"===i?"max-content":void 0,borderBottomWidth:a,children:t.map(e=>(0,n.jsx)(r.q,{label:e.label,"data-testid":e.label,isDisabled:e.isDisabled,fontSize:l.fontSize},e.label))})})}},58665:function(e,t,i){"use strict";var n=i(24246),s=i(96013);t.Z=e=>{let{...t}=e;return(0,n.jsx)(s.kCb,{boxSize:"full",align:"center",justify:"center",children:(0,n.jsx)(s.$jN,{color:"primary",...t})})}},13216:function(e,t,i){"use strict";i.r(t),i.d(t,{default:function(){return _}});var n=i(24246),s=i(96013),r=i(86677),a=i(27378),l=i(44296),d=e=>{let{data:t,...i}=e;return(0,n.jsx)(s.mQc,{colorScheme:"complimentary",...i,children:(0,n.jsx)(s.nPR,{children:t.map(e=>(0,n.jsx)(s.x45,{px:0,"data-testid":"tab-panel-".concat(e.label),children:e.content},e.label))})})},o=i(11574),c=i(11596),u=i(58665),h=i(60136),x=i(62528),m=i(43124),f=i(87162),j=i(37541),y=i(96878),b=i(22153),p=i(62711),g=i(32141),v=i(54507),_=()=>{let e=(0,r.useRouter)(),t=(0,l.T)(),i="";e.query.id&&(i=Array.isArray(e.query.id)?e.query.id[0]:e.query.id);let{data:_,isLoading:w}=(0,b.rn)(i,{skip:!i}),{error:C,isLoading:S}=(0,y.Rd)(),{tcf:Z}=(0,c.hz)(),{flags:k}=(0,c.Vb)(),z=k.dataDiscoveryAndDetection,N=(0,l.C)(p.gU);(0,a.useEffect)(()=>{if(t((0,b.db)(_)),_){let e=Z&&!!_.vendor_id&&(0,h.cj)(_.vendor_id)===h.c6.GVL;t((0,p.gQ)(e))}else(0,p.gQ)(!1)},[_,t,Z]);let{tabData:E,tabIndex:q,onTabChange:A}=(0,v.Z)({isCreate:!1});return(w||S)&&!C?(0,n.jsx)(m.Z,{title:"Systems",children:(0,n.jsx)(u.Z,{})}):(0,n.jsxs)(m.Z,{title:"System inventory",children:[(0,n.jsx)(j.Z,{heading:"System inventory",breadcrumbItems:[{title:"All systems",href:f.So},{title:(null==_?void 0:_.name)||""}],children:(0,n.jsxs)(s.xuv,{position:"relative",children:[(0,n.jsx)(o.Z,{data:E,"data-testid":"system-tabs",index:q,isLazy:!0,isManual:!0,onChange:A,width:"full",border:"full-width"}),z&&(0,n.jsxs)(s.wpx,{size:"small",className:"absolute right-2 top-2","data-testid":"integration-page-btn",onClick:()=>e.push(f.KH),children:[(0,n.jsx)(s.xvT,{children:"Integrations"}),(0,n.jsx)(x.A5,{marginLeft:2})]})]})}),N?(0,n.jsx)(g.Z,{}):null,_||w||S?(0,n.jsx)(s.gCW,{alignItems:"stretch",flex:"1",gap:"18px",maxWidth:"70vw",children:(0,n.jsx)(d,{data:E,index:q,isLazy:!0,isManual:!0,onChange:A})}):(0,n.jsxs)(s.xvT,{"data-testid":"system-not-found",children:["Could not find a system with id ",i]})]})}}},function(e){e.O(0,[8033,5559,4554,2866,9278,4833,8540,5973,8141,2017,5318,9659,2888,9774,179],function(){return e(e.s=8828)}),_N_E=e.O()}]);