"""add 'hidden' column to systems

Revision ID: abc53a00755b
Revises: 5a4859f74832
Create Date: 2024-11-11 20:18:27.770078

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "abc53a00755b"
down_revision = "5a4859f74832"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "ctl_systems",
        sa.Column("hidden", sa.BOOLEAN(), server_default="f", nullable=False),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("ctl_systems", "hidden")
    # ### end Alembic commands ###
