(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[1051],{55342:function(e,n,o){(window.__NEXT_P=window.__NEXT_P||[]).push(["/settings/domain-records",function(){return o(58316)}])},43124:function(e,n,o){"use strict";o.d(n,{Z:function(){return m}});var r=o(24246),i=o(96013),t=o(88038),s=o.n(t),a=o(86677);o(27378);var c=o(11596),d=o(72247),l=o(87162),u=()=>{let e=(0,a.useRouter)();return(0,r.jsx)(i.xuv,{bg:"gray.50",border:"1px solid",borderColor:"blue.400",borderRadius:"md",justifyContent:"space-between",p:5,mb:5,mt:5,children:(0,r.jsxs)(i.xuv,{children:[(0,r.jsxs)(i.Kqy,{direction:{base:"column",sm:"row"},justifyContent:"space-between",children:[(0,r.jsx)(i.xvT,{fontWeight:"semibold",children:"Configure your storage and messaging provider"}),(0,r.jsx)(i.wpx,{onClick:()=>{e.push(l.fz)},children:"Configure"})]}),(0,r.jsxs)(i.xvT,{children:["Before Fides can process your privacy requests we need two simple steps to configure your storage and email client."," "]})]})})},m=e=>{let{children:n,title:o,padded:t=!0,mainProps:l}=e,m=(0,c.hz)(),h=(0,a.useRouter)(),x="/privacy-requests"===h.pathname||"/datastore-connection"===h.pathname,f=!(m.flags.privacyRequestsConfiguration&&x),{data:p}=(0,d.JE)(void 0,{skip:f}),{data:v}=(0,d.PW)(void 0,{skip:f}),g=m.flags.privacyRequestsConfiguration&&(!p||!v)&&x;return(0,r.jsxs)(i.kCb,{"data-testid":o,direction:"column",h:"100vh",children:[(0,r.jsxs)(s(),{children:[(0,r.jsxs)("title",{children:["Fides Admin UI - ",o]}),(0,r.jsx)("meta",{name:"description",content:"Privacy Engineering Platform"}),(0,r.jsx)("link",{rel:"icon",href:"/favicon.ico"})]}),(0,r.jsxs)(i.kCb,{as:"main",direction:"column",py:t?6:0,px:t?10:0,h:t?"calc(100% - 48px)":"full",flex:1,minWidth:0,overflow:"auto",...l,children:[g?(0,r.jsx)(u,{}):null,n]})]})}},58316:function(e,n,o){"use strict";o.r(n);var r=o(24246),i=o(92222),t=o(59003),s=o(96013),a=o(27378),c=o(71288),d=o(43124),l=o(8540),u=o(96878);let m=(0,i.Cl)();n.default=()=>{let e=(0,a.useMemo)(()=>[m.accessor(e=>e.hostName,{header:"Name",cell:e=>(0,r.jsx)(l.G3,{value:e.getValue()})}),m.accessor(e=>e.type,{header:"Type",cell:e=>(0,r.jsx)(l.G3,{value:e.getValue()})}),m.accessor(e=>e.data,{header:"Value",cell:e=>(0,r.jsx)(l.G3,{value:e.getValue()})}),m.display({id:"actions",cell:e=>{let{row:n}=e;return(0,r.jsx)(c.Z,{copyText:n.original.data,type:"default",size:"small"})},header:"Actions",maxSize:65})],[]),{data:n,isLoading:o}=(0,u.Vh)(),h=(0,a.useMemo)(()=>(null==n?void 0:n.domain_verification_records)?n.domain_verification_records.map(e=>({hostName:"www",type:"CNAME",data:e})):[],[n]),x=(0,t.b7)({getCoreRowModel:(0,i.sC)(),columns:e,data:h,columnResizeMode:"onChange"});return(0,r.jsx)(d.Z,{title:"Domain records",children:(0,r.jsxs)(s.xuv,{"data-testid":"domain-records",children:[(0,r.jsx)(s.X6q,{marginBottom:4,fontSize:"2xl",children:"Domain records"}),(0,r.jsxs)(s.xuv,{maxWidth:"600px",children:[(0,r.jsx)(s.xvT,{marginBottom:2,fontSize:"md",children:"Set the following record on your DNS provider to continue."}),(0,r.jsxs)(s.xvT,{mb:10,fontSize:"sm",children:["Please visit"," ",(0,r.jsx)(s.rUS,{color:"complimentary.500",href:"https://fid.es/manage-dns",isExternal:!0,children:"docs.ethyca.com"})," ","for more information on how to configure Domain records."]}),o?(0,r.jsx)(s.xuv,{p:2,borderWidth:1,children:(0,r.jsx)(l.I4,{rowHeight:26,numRows:5})}):(0,r.jsx)(l.ZK,{tableInstance:x})]})]})})}}},function(e){e.O(0,[8033,5559,8540,2888,9774,179],function(){return e(e.s=55342)}),_N_E=e.O()}]);