"""data protection impact assessments additional column on systems

Revision ID: a2b9b156ebaf
Revises: adad3be6af08
Create Date: 2022-04-01 18:42:58.961132

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a2b9b156ebaf"
down_revision = "adad3be6af08"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "systems",
        sa.Column("data_protection_impact_assessment", sa.JSON(), nullable=True),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("systems", "data_protection_impact_assessment")
    # ### end Alembic commands ###
